﻿/*
  PwGenBaliktad Plugin
  Copyright (C) 2011-2012 Dominik Reichl <dominik.reichl@t-online.de>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

using KeePassLib;
using KeePassLib.Cryptography;
using KeePassLib.Cryptography.PasswordGenerator;
using KeePassLib.Security;

namespace PwGenBaliktad
{
	public sealed class BaliktadGen : CustomPwGenerator
	{
		private const string m_strSpecialCharSet =
			@"~!@#$%^&*()-_+={}[]|;:/?.,<>'`" + "\"";

		private static readonly PwUuid m_uuid = new PwUuid(new byte[] {
			0x53, 0x81, 0x36, 0xCE, 0xAE, 0xFC, 0x48, 0x3F,
			0x9E, 0x90, 0xA4, 0x4F, 0x1A, 0xF0, 0x58, 0x96 });
		public override PwUuid Uuid
		{
			get { return m_uuid; }
		}

		public override string Name
		{
			get { return "Baliktad"; }
		}

		public override ProtectedString Generate(PwProfile prf,
			CryptoRandomStream crsRandomSource)
		{
			if(prf == null) { Debug.Assert(false); }
			else
			{
				Debug.Assert(prf.CustomAlgorithmUuid == Convert.ToBase64String(
					m_uuid.UuidBytes, Base64FormattingOptions.None));
			}

			PwProfile prfSub = new PwProfile();
			prfSub.GeneratorType = PasswordGeneratorType.CharSet;
			prfSub.CharSet = new PwCharSet(PwCharSet.UpperCase + PwCharSet.LowerCase +
				PwCharSet.Digits + m_strSpecialCharSet);
			prfSub.CollectUserEntropy = false;
			prfSub.ExcludeCharacters = string.Empty;
			prfSub.ExcludeLookAlike = false;
			prfSub.Length = 11;
			prfSub.NoRepeatingCharacters = false;

			string strPw;
			while(true)
			{
				ulong u = crsRandomSource.GetRandomUInt64();
				if(u >= 18446744073709551610UL) continue;
				u %= 10UL;

				strPw = new string((char)('0' + (int)u), 1);

				byte[] pbEntropy = crsRandomSource.GetRandomBytes(32);
				ProtectedString psSub;
				if(PwGenerator.Generate(out psSub, prfSub, pbEntropy, null) !=
					PwgError.Success) { Debug.Assert(false); continue; }

				strPw += psSub.ReadString();
				if(Validate(strPw)) break;
			}

			Debug.Assert(strPw.Length == 12);
			return new ProtectedString(false, strPw);
		}

		private static bool Validate(string strPw)
		{
			string strAlpha = PwCharSet.UpperCase + PwCharSet.LowerCase;
			bool bAlpha = false;
			// bool bDigits = false; // First character is digit
			bool bSpecial = false;

			Dictionary<char, int> dict = new Dictionary<char, int>();

			foreach(char ch in strPw)
			{
				if(strAlpha.IndexOf(ch) >= 0) bAlpha = true;
				// if(PwCharSet.Digits.IndexOf(ch) >= 0) bDigits = true;
				if(m_strSpecialCharSet.IndexOf(ch) >= 0) bSpecial = true;

				if(dict.ContainsKey(ch)) dict[ch] = dict[ch] + 1;
				else dict.Add(ch, 1);
			}

			bool bRepeated = false;
			foreach(KeyValuePair<char, int> kvp in dict)
			{
				if(kvp.Value > 2) return false;
				if(kvp.Value == 2)
				{
					if(bRepeated) return false;
					bRepeated = true;
				}
			}

			if(!bAlpha) return false;
			// if(!bDigits) return false;
			if(!bSpecial) return false;
			return true;
		}
	}
}
